/* capilayer.c - a translation layer between the various CAPI versions and
 *		 ctime.c
 *
 * $Id: capilayer.c,v 1.13 2005/02/09 01:24:28 ianmacd Exp $
 *
 * Author: Ian Macdonald <ian@caliban.org>
 *
 * Copyright (C) 2004-2005 Ian Macdonald
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2, or (at your option)
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software Foundation,
 *   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <ruby.h>


/* Required for ctapi.h */
#ifdef WIN32
#define CAPI_FUNC(a)  __declspec(dllimport) a
#else
#define CAPI_FUNC(a) a
#endif

#include <ctapi.h>
#include "capilayer.h"


/*
 * capi_connect
 *
 * Connects and authenticates to the server.
 */
CAPIStatus
capi_connect (const char *in_host,
	      const char *in_user,
	      const char *in_passwd, CAPISession * out_session)
{
  CAPIStatus stat;

#if defined(HAVE_CAPI_20) || defined(HAVE_CAPI_25)
  stat = CAPI_Connect (in_host, CAPI_FLAG_NONE, out_session);
  if (stat != CAPI_STAT_OK)
    {
      return stat;
    }

  stat = CAPI_Logon (in_user, in_passwd, "", CAPI_FLAG_NONE, out_session);
#else
  stat = CSDK_CreateSession (CSDK_FLAG_NONE, out_session);
  if (stat != CAPI_STAT_OK)
    {
      return stat;
    }

  stat = CSDK_Connect (*out_session, CSDK_FLAG_NONE, in_host);
  if (stat != CAPI_STAT_OK)
    {
      return stat;
    }

  stat = CSDK_Authenticate (*out_session, CSDK_FLAG_NONE, in_user, in_passwd);
#endif

  return stat;
}


void
capi_get_status_string (CAPIStatus status, const char **out_string)
{
#if defined(HAVE_CAPI_20) || defined(HAVE_CAPI_25)
  CAPI_GetStatusString (status, out_string);

#else
  CSDK_GetStatusString (status, out_string);
#endif
}


CAPIStatus
capi_get_capabilities (CAPISession in_session,
		       CAPICapabilityID in_capabilityID,
		       const char **out_value)
{
#if defined(HAVE_CAPI_20) || defined(HAVE_CAPI_25)
  return CAPI_GetCapabilities (in_session, in_capabilityID, CAPI_FLAG_NONE,
			       out_value);
#else
  return CSDK_GetCapabilities (in_session, in_capabilityID, CSDK_FLAG_NONE,
			       out_value);
#endif
}


CAPIStatus
capi_destroy_multiple_handles (CAPISession in_session,
			       CAPIHandle * in_handle, int num_handles)
{
#if defined(HAVE_CAPI_20) || defined(HAVE_CAPI_25)
  return CAPI_DestroyHandles (in_session, in_handle, num_handles,
			      CAPI_FLAG_NONE);
#else
  return CSDK_DestroyMultipleHandles (in_session, in_handle, num_handles,
				      CSDK_FLAG_NONE);

#endif
}


/*
 * capi_disconnect
 *
 * Logs out of the server and disconnects.
 * Invalidates the session too.
 */
CAPIStatus
capi_disconnect (CAPISession * in_session)
{
#if defined(HAVE_CAPI_20) || defined(HAVE_CAPI_25)
  return CAPI_Logoff (in_session, CAPI_FLAG_NONE);
#else
  CSDK_Deauthenticate (*in_session, CSDK_FLAG_NONE);

  CSDK_Disconnect (*in_session, CSDK_FLAG_NONE);

  return CSDK_DestroySession (in_session);
#endif
}


CAPIStatus
capi_get_handle (CAPISession in_session,
		 const char *in_user, CAPIHandle * out_handle)
{
#if defined(HAVE_CAPI_20) || defined(HAVE_CAPI_25)
  return CAPI_GetHandle (in_session, in_user, CAPI_FLAG_NONE, out_handle);
#else
  return CSDK_GetHandle (in_session, in_user, CSDK_FLAG_NONE, out_handle);
#endif
}


CAPIStatus
capi_get_handle_info (CAPISession in_session,
		      CAPIHandle in_handle,
		      CAPIFlag in_flags, const char **out_info)
{
#if defined(HAVE_CAPI_20)
  return CAPI_HandleInfo (in_session, in_handle, in_flags,
			  (char **) out_info);
#elif defined(HAVE_CAPI_25)
  return CAPI_HandleInfo (in_session, in_handle, in_flags, out_info);
#else
  return CSDK_GetHandleInfo (in_session, in_handle, in_flags, out_info);
#endif
}


int
read_callback (void *in_userdata,
	       char *in_data, size_t in_datasize, size_t * out_datasize)
{

  VALUE out_rb_str = (VALUE) in_userdata;

  rb_ctime_debug ("In read_callback...\n");

  rb_ctime_debug ("Got value...\n");

  rb_str_cat (out_rb_str, in_data, in_datasize);

  rb_ctime_debug ("Did str_cat...\n");

  *out_datasize = in_datasize;

  rb_ctime_debug ("Exiting read_callback...\n");

  return CAPI_CALLBACK_CONTINUE;
}


/*
 * This version was created in an attempt to debug strange memory failures in
 * the CAPI 9 version. Unfortunately, it didn't help. Keeping in case it
 * becomes useful again.
 */
CAPIStatus
capi_create_read_callback_stream (CAPISession in_session,
				  CAPIStream * out_stream, VALUE out_rb_str)
{
#if defined(HAVE_CAPI_20)
  return CAPI_CreateCallbackStream (in_session, out_stream, NULL,
				    NULL, read_callback, (void *) out_rb_str,
				    CAPI_FLAG_NONE);
#elif defined(HAVE_CAPI_25)
  return CAPI_CreateCallbackStream (in_session, out_stream, NULL,
				    NULL, read_callback, (void *) out_rb_str,
				    CAPI_FLAG_NONE);
#else
  return CSDK_CreateCallbackStream (in_session, out_stream, NULL, NULL,
				    read_callback, (void *) out_rb_str,
				    CSDK_FLAG_NONE);
#endif
}


CAPIStatus
capi_create_memory_stream (CAPISession in_session,
			   CAPIStream * out_stream,
			   const char *in_buffer, const char **out_buffer)
{
#if defined(HAVE_CAPI_20)
  return CAPI_CreateMemoryStream (in_session, out_stream, in_buffer,
				  (char **) out_buffer, CAPI_FLAG_NONE);
#elif defined(HAVE_CAPI_25)
  return CAPI_CreateMemoryStream (in_session, out_stream, in_buffer,
				  out_buffer, CAPI_FLAG_NONE);
#else
  return CSDK_CreateMemoryStream (in_session, out_stream, in_buffer,
				  out_buffer, CSDK_FLAG_NONE);
#endif
}


CAPIStatus
capi_destroy_stream (CAPISession in_session, CAPIStream * in_stream)
{
#if defined(HAVE_CAPI_20) || defined(HAVE_CAPI_25)
  return CAPI_DestroyStreams (in_session, in_stream, 1, CAPI_FLAG_NONE);
#else
  return CSDK_DestroyStream (in_session, in_stream);
  /* return CSDK_DestroyMultipleStreams (in_session, in_stream, 1,
     CSDK_FLAG_NONE); */
#endif
}


/* capi_fetch_event_by_uid */
CAPIStatus
capi_fetch_event_by_uid (CAPISession in_session,
			 CAPIHandle in_handle,
			 const char *uid,
			 const char **in_req_properties,
			 int num_properties, CAPIStream in_stream)
{
  CAPIStatus status;

#if defined(HAVE_CAPI_20)
  status = CAPI_FetchEventByID (in_session, in_handle,
				CAPI_FLAG_NONE, (char *) uid, NULL,
				CAPI_THISINSTANCE,
				(char **) in_req_properties,
				num_properties, in_stream);
#elif defined(HAVE_CAPI_25)
  status = CAPI_FetchEventByID (in_session, in_handle,
				CAPI_FLAG_NONE, uid, NULL,
				CAPI_THISINSTANCE, in_req_properties,
				num_properties, in_stream);
#else
  char const *uidset[2];
  uidset[0] = (char const *) uid;
  uidset[1] = NULL;

  status = CSDK_FetchEventsByUID (in_session, CSDK_FLAG_NONE,
				  in_handle, uidset, NULL,
				  CAPI_THISINSTANCE, in_req_properties,
				  in_stream, NULL);
#endif

  return status;
}


/*
 * capi_fetch_events_by_range
 *
 * WARNING: version 9 of the API expects in_handles and in_req_properties to
 * be null terminated. It's wise to always null-terminate these arrays unless
 * you know exactly which version of the API is being used.
 */
CAPIStatus
capi_fetch_events_by_range (CAPISession in_session,
			    CAPIHandle * in_handles,
			    int num_handles,
			    const char *in_start,
			    const char *in_end,
			    const char **in_req_properties,
			    int num_properties, CAPIStream in_stream)
{
  CAPIStatus status;

#if defined(HAVE_CAPI_20)
  CAPIStatus *stats;
  stats = calloc (num_handles, sizeof (CAPIStatus));

  status = CAPI_FetchEventsByRange (in_session, in_handles, num_handles,
				    stats, CAPI_FLAG_NONE,
				    (char *) in_start,
				    (char *) in_end,
				    (char **) in_req_properties,
				    num_properties, in_stream);
  free (stats);
#elif defined(HAVE_CAPI_25)
  CAPIStatus *stats;
  stats = calloc (num_handles, sizeof (CAPIStatus));

  status = CAPI_FetchEventsByRange (in_session, in_handles, num_handles,
				    stats, CAPI_FLAG_NONE, in_start,
				    in_end, in_req_properties,
				    num_properties, in_stream);

  free (stats);
#else
  status = CSDK_FetchEventsByRange (in_session, CSDK_FLAG_NONE,
				    in_handles, in_start, in_end,
				    in_req_properties, in_stream, NULL);
#endif

  return status;
}


CAPIStatus
capi_set_identity (CAPISession in_session, const char *in_user)
{
#if defined(HAVE_CAPI_20)
  return CAPI_SetIdentity (in_user, CAPI_FLAG_NONE, in_session);
#elif defined(HAVE_CAPI_25)
  return CAPI_SetIdentity (in_session, in_user, CAPI_FLAG_NONE);
#else
  return CSDK_SetIdentity (in_session, in_user, CSDK_FLAG_NONE);
#endif
}


/*
 * capi_store_event
 *
 * This function is used to store a single event inside calendar. It assumes
 * only one event is in the stream passed in, and returns the event id in the
 * passed in parameter.
 *
 * WARNING: The CAPI version 9 changed behaviour. It now assumes the attendees
 * are listed in the iCalendar stream. There's no way to make this transparent.
 * You'll need to know which version the API you're using and either open the
 * correct agendas or place them in the iCalendar stream.
 */
CAPIStatus
capi_store_event (CAPISession in_session,
		  CAPIHandle * in_handles,
		  int num_handles, CAPIStream in_stream,
		  char *uid_buf, int uid_buflen, int replace_flags)
{
  CAPIStatus status;
  const char *new_uid = NULL;

#if defined(HAVE_CAPI_20) || defined(HAVE_CAPI_25)
  CAPIStatus *stats;
  stats = calloc (num_handles, sizeof (CAPIStatus));

  status = CAPI_StoreEvent (in_session, in_handles, num_handles, stats,
			    CAPI_STORE_UPDATE, in_stream);

  if (NULL != uid_buf)
    CAPI_GetLastStoredUID (in_session, &new_uid, CAPI_FLAG_NONE);

  if (NULL != new_uid && NULL != uid_buf)
    {
      strncpy (uid_buf, new_uid, uid_buflen);
      new_uid = NULL;
    }

  free (stats);
#else
  CSDKRequestResult req_result = NULL;

  if (replace_flags & RBCTIME_REPLY)
    status = CSDK_StoreEvents (in_session, CSDK_FLAG_STORE_REPLY, in_stream,
			       &req_result);
  else
    status = CSDK_StoreEvents (in_session, CSDK_FLAG_STORE_IMPORT, in_stream,
			       &req_result);

  if (replace_flags & RBCTIME_MODIFY)
    status = CSDK_StoreEvents (in_session, CSDK_FLAG_STORE_MODIFY,
			       in_stream, &req_result);

  /* Failed or we have no uid buffer; either way, we don't need the uid */
  if (status != CAPI_STAT_OK || NULL == uid_buf)
    return status;

  CSDK_GetFirstResult (req_result, NULL, &new_uid, &status);

  if (replace_flags & RBCTIME_REPLACE &&
      CAPI_STAT_OK == status && NULL != uid_buf && NULL != new_uid)
    capi_delete_event (in_session, in_handles, num_handles, uid_buf);

  if (NULL != new_uid)
    strncpy (uid_buf, new_uid, uid_buflen);

  /* After calling this, new_uid is invalid */
  CSDK_DestroyResult (&req_result);

#endif

  return status;
}


/*
 * capi_delete_event
 *
 * Deletes event with the given uid. No support for repeating events.
 */
CAPIStatus
capi_delete_event (CAPISession in_session,
		   CAPIHandle * in_handles, int num_handles, const char *uid)
{
  CAPIStatus status;

#if defined(HAVE_CAPI_20) || defined(HAVE_CAPI_25)
  CAPIStatus *stats;
  stats = calloc (num_handles, sizeof (CAPIStatus));

  status = CAPI_DeleteEvent (in_session, in_handles, num_handles, stats,
			     CAPI_FLAG_NONE, uid, NULL, 0);

  free (stats);
#else
  char const *uidset[2];
  uidset[0] = (char const *) uid;
  uidset[1] = NULL;

  status = CSDK_DeleteEvents (in_session, CSDK_FLAG_NONE, uidset,
			      NULL, CAPI_THISINSTANCE, NULL);
#endif

  return status;
}
