# $Id: event.rb,v 1.26 2005/03/15 23:59:54 ianmacd Exp $

require "ctime/attendee"
require "date"

class CTime
  class Agenda
    class Event

      attr_accessor :uid, :attendees, :organizer, :ics,
		    :summary, :description, :location
      alias_method :organiser, :organizer
      alias_method :organiser=, :organizer=

      def initialize(start_time, duration, summary, location, description)
        @attendees = Array.new
        @start_time = start_time
	@prev_start_time = nil
        @end_time = nil
        @duration = duration
        @summary = Event.unescape_string(summary) if summary
        @location = Event.unescape_string(location) if location
        @description = Event.unescape_string(description) if description
        @ics = nil
        @uid = nil
	@organizer = nil
      end


      def start_time=(start_time)
	@prev_start_time = start_time_str

        if start_time.kind_of? String
          @start_time = Event.parse_date_time(start_time)
        elsif start_time.respond_to? :gmtime
          # probably a Time object
          @start_time = start_time.gmtime
        elsif start_time.respond_to? :new_offset
          # probably a DateTime object
          @start_time = start_time.new_offset(0)
        else
	  @prev_start_time = nil
        end

	# nothing changed
	@prev_start_time = nil if @prev_start_time == start_time_str
      end
      alias_method :start=, :start_time=


      def start_time
        if @start_time.kind_of? String
          @start_time
        elsif @start_time.respond_to? :gmtime
          @start_time.gmtime
        elsif @start_time.respond_to? :new_offset
          @start_time = @start_time.new_offset(0)
        else
          nil
        end
      end
      alias_method :start, :start_time


      def start_time_str
        if @start_time.kind_of? String
          @start_time
        elsif @start_time.respond_to? :strftime
          start_time.strftime(CTime::STRFTIME)
        else
          nil
        end
      end


      def end_time=(end_time)
        if end_time.kind_of? String
          @end_time = Event.parse_date_time(end_time)
        elsif end_time.respond_to? :gmtime
          # probably a Time object
          @end_time = end_time.gmtime
        elsif end_time.respond_to? :new_offset
          # probably a DateTime object
          @end_time = end_time.new_offset(0)
        else
          nil
        end
      end
      alias_method :end=, :end_time=


      def end_time
        if @end_time

          if @end_time.respond_to? :gmtime
	    @end_time = @end_time.gmtime
          elsif @end_time.respond_to? :new_offset
	    @end_time = @end_time.new_offset(0)
          end

          return @end_time

        elsif @start_time && @duration
          if @start_time.kind_of? DateTime
	    # DateTime object + 1 adds one day
            @end_time = @start_time + @duration / 86400.0
          else
	    # Time object + 1 adds one second
            @end_time = @start_time + @duration
          end
        elsif @start_time
          @end_time = @start_time
        else
          nil
        end
      end
      alias_method :end, :end_time


      def end_time_str
        if @end_time.kind_of? String
          @end_time
        elsif @end_time.respond_to? :strftime
          end_time.strftime(CTime::STRFTIME)
        else
          nil
        end
      end


      def duration=(duration)
        if duration.kind_of? String
          @duration = Event.parse_duration(duration)
        elsif duration.kind_of? Integer
          @duration = duration
        else
          nil
        end
      end


      #
      # Returns the duration as an integer (in number of seconds)
      #
      def duration
        if @duration
          @duration
        elsif @start_time && @end_time
          @duration = (@end_time - @start_time).to_i
        else
          nil
        end
      end


      #
      # Returns the duration as a String in iCalendar format
      #
      def duration_str
        dur = duration
        days, dur = dur.divmod(24 * 3600)
        hours, dur = dur.divmod(3600)
        mins, secs = dur.divmod(60)

        return "P%02dDT%02dH%02dM%02dS" % [days, hours, mins, secs]
      end


      #
      # Escape newlines, commas, tabs, as per RFC 2445
      #
      def description_str
	if @description.nil?
	  return ""
	end

	# Unescape first to ensure cleanliness (i.e. no multiple \)
	return Event.escape_string(Event.unescape_string(@description))
      end


      #
      # Escape newlines, commas, tabs, as per RFC 2445
      #
      def location_str
	if @location.nil?
	  return ""
	end

	# Unescape first to ensure cleanliness (i.e. no multiple \)
	return Event.escape_string(Event.unescape_string(@location))
      end


      #
      # Escape newlines, commas, tabs, as per RFC 2445
      #
      def summary_str
	if @summary.nil?
	  return ""
	end

	# Unescape first to ensure cleanliness (i.e. no multiple \)
	return Event.escape_string(Event.unescape_string(@summary))
      end


      #
      # Return a human-readable representation of the event.
      #
      def to_s
        result = "Event:\n" +
                 "\tuid=#{uid()}\n" +
                 "\tsummary=#{summary()}\n" +
                 "\tlocation=#{location()}\n" +
                 "\tdescription=#{description()}\n" +
                 "\tstart_time=#{start_time()}\n" +
                 "\tend_time=#{end_time()}\n" +
                 "\tduration=#{duration()}\n"
        @attendees.each { |attendee| result << attendee.inspect }

        result
      end


      #
      # Return the raw iCalendar stream for the event.
      #
      def inspect
        @ics
      end


      #
      # Parses an iCalendar Date-Time value. Returns a Ruby Time object.
      #
      # Currently it only recognises Date with UTC and Date with Local Time
      # formats.
      #
      def Event.parse_date_time(date_time_str)
        if date_time_str =~ /^(\d{4})(\d\d)(\d\d)T(\d\d)(\d\d)(\d\d)Z$/
          Time.gm($1, $2, $3, $4, $5, $6).localtime
        elsif date_time_str =~ /^(\d{4})(\d\d)(\d\d)T(\d\d)(\d\d)(\d\d)$/
          Time.local($1, $2, $3, $4, $5, $6)
        else
          nil
        end
      end


      #
      # Parses an iCalendar Date value. Returns a Ruby Time (in localtime).
      #
      # end_of_day determines if the time object should be at the very end of
      # the day. By default it's at the very beginning of the day.
      #
      def Event.parse_date(date_str, end_of_day = false)
        if date_str =~ /^(\d{4})(\d\d)(\d\d)$/
          if end_of_day
            Time.local($1, $2, $3, 23, 59, 59, 999999)
          else
            Time.local($1, $2, $3, 0, 0, 0, 0)
          end
        else
          return nil
        end
      end


      #
      # Parses an iCalendar Duration value. Returns an integer number of
      # seconds.
      #
      def Event.parse_duration(duration)
        weeks = 0
        days = 0
        hours = 0
        mins = 0
        secs = 0

        duration.sub!(/^[-+]*P/, "")

        duration.sub!(/^(\d+)W/) do |match|
          weeks = $1.to_s.to_i
          ""
        end

        duration.sub!(/^(\d+)D/) do |match|
          days = $1.to_s.to_i
          ""
        end

        duration.sub!(/^T/, "")

        duration.sub!(/^(\d+)H/) do |match|
          hours = $1.to_s.to_i
          ""
        end

        duration.sub!(/^(\d+)M/) do |match|
          mins = $1.to_s.to_i
          ""
        end

        duration.sub!(/^(\d+)S/) do |match|
          secs = $1.to_s.to_i
          ""
        end

        return (weeks * 7 * 24 * 3600) +
               (days * 24 * 3600) +
               (hours * 3600) +
               (mins * 60) +
               secs
      end


      #
      # Unescapes a string, converting \n to 0xA, \t to 0x9, etc.
      # This is used to decode the Description and other fields, which
      # come over thus encoded.
      #
      # This method only handles simple one character encoding, not any of the 
      # \nnn (octal), \xnn( (hex), \cx, etc multi-character encodings.
      #
      def Event.unescape_string(escaped_string)
        in_escape = false

        unescaped_array = escaped_string.split(//).map do |char|
          if in_escape
            in_escape = false

            case char
            when "n" then "\n"
            when "r" then "\r"
	    when "\\" then ""
            else char
            end
          else
            case char
            when "\\"
              in_escape = true
              ""
            else
              char
            end
          end
        end

        unescaped_array.to_s
      end


      #
      # Escapes a string, converting 0xA to \n, to 0x9 \t, etc.
      # This is used to encode the Description and other fields, which
      # come over thus encoded, so we send the same back.
      #
      # This method only handles simple one character encoding, not any of the 
      # \nnn (octal), \xnn( (hex), \cx, etc multi-character encodings.
      #
      def Event.escape_string(unescaped_string)

        escaped_array = unescaped_string.split(//).map do |char|
          case char
          when "\r" then ""
          when "\n" then "\\n"
          when "," then "\\,"
          else char
          end
        end

        escaped_array.to_s
      end


      def Event.parse_events(raw_event_string)
        events = Array.new
        inside_event = false
        inside_alarm = false
        name = ""
        param_hash = {}
        value = ""
        event = nil

        if $DEBUG
          $stderr.puts "Raw string:"
          $stderr.puts raw_event_string
          $stderr.puts "---------"
        end

        #
        # Get rid of folded lines.
        #
        # I don't see where this is in the spec, but Steltor also
        # "continues" lines by ending them with =\r\n
        #
        # It's in the VCard spec, not iCalendar; I guess some apps are
        # braindead enough to use that folding convention for iCalendar.
        #
        raw_event_string.gsub!(/\r\n[ \t]/, "")
        raw_event_string.gsub!(/=\r\n/, "")

        raw_event = nil
        raw_event_string.each_line do |line|
          (name, param_hash, value) = Event.parse_line(line)
          next if name == ""

          if $DEBUG
            $stderr.puts "Name: #{name}   Value: #{value}"
            param_hash.each_key do |key|
              $stderr.puts "\t#{key}=#{param_hash[key]}"
            end
          end

          #
          # Parse out events
          #
          if ! inside_event
            if name == "BEGIN" && value == "VEVENT"
              inside_event = true
              event = Event.new(nil, nil, nil, nil, nil)
              events.push(event)
	      raw_event = [line]
            end
          elsif name == "BEGIN" && value == "VALARM"
	    inside_alarm = true
	    raw_event << line
	    next
          elsif name == "END" && value == "VALARM"
            inside_alarm = false
	    raw_event << line
	    next
          elsif inside_alarm
	    raw_event << line
	    next
          else
	    raw_event << line
            case name
            when "UID"
              event.uid = value
            when "DTSTART"
              if param_hash["VALUE"] == "DATE"
                event.start_time = Event.parse_date(value)
	      else
		event.start_time = value
              end
            when "DTEND"
	      if param_hash["VALUE"] == "DATE"
		event.end_time = Event.parse_date(value, true)
	      else
		event.end_time = value
	      end
            when "DURATION"
              event.duration = value
            when "SUMMARY"
              event.summary = Event.unescape_string(value)
            when "LOCATION"
              event.location = Event.unescape_string(value)
            when "DESCRIPTION"
              event.description = Event.unescape_string(value)
            when "ATTENDEE"
              event.attendees.push(Attendee.parse_attendee(param_hash, value))
            when "ORGANIZER"
              event.organizer = value.sub(/^mailto:/i, "")
            when "END"
              if value == "VEVENT"
                inside_event = false
		event.ics = raw_event.join("\n")
                event = nil
              end
            else
              $stderr.puts "Unmatched name *#{name}*" if $DEBUG
            end
          end
        end

        return events
      end


      def Event.parse_line(line)
        name = ""
        value = ""
        param_hash = Hash.new

        line.sub!(/\r\n$/, "")

        if line =~ /^([^:]*):(.*)$/
          name = $1
          value = $2
          if name =~ /^([^;]*);(.*)$/
            name = $1
            param_string = $2.unpack("M*").to_s
            param_string.split(/;/).each do |name_value_string|
              if name_value_string =~ /^([^=]*)=(.*)$/
                param_hash[$1] = $2
              end
            end
          end

          # Decode quoted printable strings and convert from UTF-8 to Latin 1.
	  # The final #pack is needed to get a string from an array of
	  # integers.
	  begin
            name = name.unpack("M*").to_s.unpack("U*").pack("c*")
            value = value.unpack("M*").to_s.unpack("U*").pack("c*")
	  rescue ArgumentError
	    # Oracle sometimes sends us malformed UTF-8.
	    $stderr.puts "Bad UTF-8 in #{name}:\n#{value.inspect}" if $DEBUG
	  end
        end

        return [name, param_hash, value]
      end

    end
  end
end
